import LoginService from '../service/login-service.js';
import regeneratorRuntime from '../../libs/regenerator-runtime/index.js';
import { toPromiseWX, setStorageValueForKey } from '../index.js';

const login = new LoginService();
const loginPromise = toPromiseWX('login');

const saveLoginData = (loginData) => {
  const app = getApp();
  setStorageValueForKey('open_id', loginData.openid);
  setStorageValueForKey('srd_session', loginData.srd_session);
  if (loginData && loginData.unionid) {
    setStorageValueForKey('union_id', loginData.unionid);
    app && app.setWXUnionID(loginData.unionid);
  }
}

export const wxLogin = async () => {
  try {
    // 调用 wx.login
    const res = await loginPromise();
    // 根据上传的 code 返回 open_id 等信息
    const loginData = await login.checkLogin(res.code);
    // 本地存储 open_id srd_session union_id
    await saveLoginData(loginData);
  } catch (error) {
    // 注意: 此处应异常上报
    console.log(error);
  }
};

export const getUserInfo = async (e) => {
  const event = global.event;
  try {
    if (e.detail.errMsg === 'getUserInfo:ok') {
      await setStorageValueForKey('userInfo', e.detail.userInfo);
      const srd_session = wx.getStorageSync('srd_session') || '';
      const json = await login.decodeUserInfo(srd_session, e.detail.encryptedData, e.detail.iv);
      if (json && json.union_id) {
        wechatUnionIdIsBind(json.union_id, e.detail.userInfo);
      } else {
        throw new Error('微信授权失败，请使用手机号登录');
      }
    } else {
      throw new Error('微信授权失败，请使用手机号登录');
    }
  } catch (error) {
    event.emit('wechat-login-error', error);
  }
}

export const getPhoneNumber = async (e) => {
  const event = global.event;
  const app = getApp();
  try {
    if (e.detail.errMsg === 'getPhoneNumber:ok') {
      const srd_session = wx.getStorageSync('srd_session') || '';
      const union_id = wx.getStorageSync('union_id') || '';
      const phoneNumberInfo = await login.decodeUserInfo(srd_session, e.detail.encryptedData, e.detail.iv);
      const { phoneNumber, countryCode } = phoneNumberInfo;
      if (!(countryCode && phoneNumber)) {
        throw new Error('手机号获取失败，请使用手机号登录');
      } 
      const bindResult = await login.bindMiniAppByAuto(union_id, phoneNumber, countryCode);
      if (bindResult && bindResult.is_register === 0) {
        const newUserInfo = {
          is_bind: bindResult.is_bind,
          mobile: bindResult.profile,
          session_key: bindResult.session_key,
          uid: bindResult.uid,
          ssouid: bindResult.ssouid,
          union_id: union_id,
          vip: bindResult.vip
        };

        let userInfo = wx.getStorageSync('userInfo');
        userInfo = Object.assign(userInfo, newUserInfo);
        await setStorageValueForKey('userInfo', userInfo);
        await setStorageValueForKey('sessionkey', bindResult.session_key);

        app.globalData.userInfo = userInfo;
        setTimeout(() => {
          event.emit('wechat-login-success');
        }, 1000);
        event.emit('user-login-success', newUserInfo);
        wx.showToast({
          icon: 'none',
          title: '手机号授权成功！\r\n为享受更多权益，已为您注册有货会员！',
          duration: 3000
        });
      } else if (bindResult && (bindResult.is_register === 2 || bindResult.is_register === 3 || bindResult.is_register === 4)) {
        const newUserInfo = {
          is_bind: bindResult.is_bind,
          mobile: bindResult.profile,
          session_key: bindResult.session_key,
          uid: bindResult.uid,
          ssouid: bindResult.ssouid,
          union_id: union_id,
          vip: bindResult.vip
        };

        let userInfo = wx.getStorageSync('userInfo');
        userInfo = Object.assign(userInfo, newUserInfo);
        await setStorageValueForKey('userInfo', userInfo);
        await setStorageValueForKey('sessionkey', bindResult.session_key);

        app.globalData.userInfo = userInfo;
        event.emit('wechat-login-success');
        event.emit('user-login-success', newUserInfo);
      }
    } else {
      throw new Error('手机号获取失败，请使用手机号登录');
    }
  } catch (error) {
    event.emit('wechat-login-error', error);
  }
}

export const checkUnionIdIsBind = (e) => {
  let union_id = wx.getStorageSync('union_id') || '';
  let userInfo = wx.getStorageSync('userInfo') || {};
  wechatUnionIdIsBind(union_id, userInfo);
}

const wechatUnionIdIsBind = async (union_id, userInfo) => {
  const event = global.event;
  let globalData = getApp().globalData;
  try {
    const user_bind_info = await login.wechatUserIsBind(union_id);
    if (user_bind_info && user_bind_info.is_bind === 'Y') {
      const { mobile, session_key, uid, ssouid } = user_bind_info;
      // 保存 uid session_key
      const newUserInfo = Object.assign(userInfo, { mobile, session_key, uid, ssouid });
      await setStorageValueForKey('userInfo', newUserInfo);
      await setStorageValueForKey('sessionkey', session_key);

      // 赋值全局变量
      globalData.sessionkey = session_key;
      globalData.userInfo = userInfo;
      // 上传保存头像
      login.sendWeChatUserData(uid, userInfo.nickName, userInfo.avatarUrl).catch(error => {});
      // 触发事件登录成功
      event.emit('wechat-login-success', newUserInfo);
      event.emit('user-login-success', newUserInfo);
    } else {
      // 修改按钮状态, 获取手机号绑定手机号
      event.emit('wechat-login-change-status', { openType: 'getPhoneNumber' });
    }
  } catch (error) {
    event.emit('wechat-login-error', error);
  }
}